  /*HEADER------------------------------------------------------------------------
|
| PROGRAM:  calc_cox_grad_hessian. .

 PURPOSE: Calculates Hessian and Gradient for iterative solution of Cox regression.
	First version:  06/01/2017, Yury Vilk 	
+------------------------------------------------------------------------HEADER*/


         %MACRO calc_cox_grad_hessian(tbl_cox_sscp_in=
																,tbl_model_coeff_in =
																,tbl_mean_vars_by_st_out=
																,tbl_resid_sum_out=
                                                                ,tbl_baseln_hazard=
																,tbl_Gradient_out =
																,tbl_Hessian_out =
	                                                         ) ;

   %PUT;
   %PUT -----   Start of the macro calc_cox_grad_hessian  iter_nb=&iter_nb. ;
   %PUT;

	 %LOCAL _new_names_str S_names Q_names S1_names d1_names S11_names S2_names;
  
	  %PUT  prefix0_dp=&prefix0_dp.;
	   %PUT  prefix_dp=&prefix_dp. ;

       /* Central side */
/*
     	%LET infolder_lib=infolder;
        %LET msoc_lib=msoc;
   	   %LET dplocal_lib=dplocal; 
	   %LET  prefix0_dp=&prefix0.;
	   %LET  prefix_dp=&prefix. ;

        %LET tbl_cox_sscp_in=dplocal.&prefix.sscp;
		%LET tbl_model_coeff_in =&infolder_lib..&prefix_prev.Model_Coeff;
        
		%LET tbl_mean_vars_by_st_out=dplocal.&prefix0.mean_vars_by_st;
		%LET tbl_resid_sum_out=dplocal.&prefix.resid_sum;
	   %LET tbl_baseln_hazard=dplocal.&prefix0.baseln_hazard
		%LET tbl_Gradient_out =dplocal.&prefix.Gradient;
		%LET tbl_Hessian_out =&dplocal_lib..&prefix.Hessian;
	     
	  */
	
					/* DP side */
	/*  
	    * %LET prefix0_l=&prefix0.;
	    * %LET prefix_l=&prefix.;		
	      %LET  prefix0_dp=&prefix0_dp.;
	      %LET  prefix0_dp=&prefix_dp. ;

		%LET tbl_cox_sscp_in=&dplocal_lib..&prefix0_dp.sscp; 
		%LET tbl_model_coeff_in =&infolder_lib..&prefix_prev.Model_Coeff;
		%LET tbl_mean_vars_by_st_out=&dplocal_lib..&prefix0_dp.mean_vars_by_st;
		%LET tbl_resid_sum =&dplocal_lib..&prefix_dp.resid_sum;
	   %LET tbl_baseln_hazard=&dplocal_lib..&prefix0.baseln_hazard
		%LET tbl_Gradient_out =&dplocal_lib..&prefix_dp.Gradient;
		%LET tbl_Hessian_out =&dplocal_lib..&prefix_dp.Hessian;
	 */

	   DATA &prefix_dp.sscp_in_srt;
	    SET  &tbl_cox_sscp_in.;
		 	_NAME_=LOWCASE(_NAME_);
		RUN;

		 PROC SORT DATA=&prefix_dp.sscp_in_srt ;
		     BY  &strata_vars. &dependent_vars. _NAME_	;
	     RUN;

		/*
	   PROC SORT DATA=&tbl_cox_sscp_in. OUT=&prefix_dp.sscp_in_srt ;
		     BY  &strata_vars. &dependent_vars. _NAME_	;
	  RUN;
	 */

	 	/* Always create the table &dplocal_lib..&prefix0.event_set_csum at the first iteration.
		   If &exist_init_events_time_set.=0  then replace the table at the iter_nb=2 because
		   it will have corrected values for cumulative weight _WEIGHT_.
	   */

  %IF &iter_nb.=1 OR (&iter_nb.=2 AND &exist_init_events_time_set.=0 
                                      AND  &pos_dp_cd_in_strata_vars. =0)  
  %THEN  
			%DO;

	  	/* Scalar _S0_ below is scalar _S_ at \betas=0 and is equal to Cum sum of weights
		  This will be used to calculate Cox log likelihood without covariates.*/

	   PROC SORT DATA=&prefix_dp.sscp_in_srt(KEEP= _NAME_ _TYPE_ &strata_vars. &dependent_vars. _WEIGHT_) 
           OUT=&dplocal_lib..&prefix0_dp.event_set_csum(DROP=_NAME_ _TYPE_ 
																  RENAME=( _WEIGHT_=_S0B_ )
                                                                    );
		     BY  &strata_vars. &dependent_vars. ;
		WHERE  LOWCASE(_NAME_)="intercept" AND _TYPE_="SSCP";
	  RUN;

	        %END;  /* End of do-group for &iter_nb.=1 OR (&iter_nb.=2 AND &exist_init_events_time_set.=0)*/

 /* Create at the first iteration a  number of necessary permanent datasets that does not 
	change with each iteration (does not depend on regr. coeff.) */

			 %IF   &iter_nb. = 1  %THEN
 				%DO;

	 DATA &dplocal_lib..&prefix0_dp.Vector_d(RENAME=(Intercept=_d_w_ _EVENT_=_d_));
	    SET &prefix_dp.sscp_in_srt	;
		WHERE     _TYPE_  IN ("SUM")  AND LOWCASE(_NAME_)="intercept" ;
		KEEP &strata_vars. &dependent_vars. Intercept &independent_vars. _EVENT_;
	RUN;

					/* Add suffix to &independent_vars names this will be useful later.*/
	%modify_colnames(tbl_name=&dplocal_lib..&prefix0_dp.Vector_d,var_lst= &independent_vars.,prefix=,suffix=_d1);
 
	         /* The stats by strata will be used to calculate survival function for each strata 
				at mean values of covariates.*/

	  DATA &dplocal_lib..&prefix0_dp.sum_vars_by_st_time(RENAME=(Intercept=_SUMWGT_));
	   * SET &prefix.sscp_c1	;
		SET &prefix_dp.sscp_in_srt;
		WHERE     _TYPE_  IN ("SUM")  AND LOWCASE(_NAME_)="intercept" ;
		KEEP  &strata_vars. &dependent_vars. Intercept &independent_vars.;
	RUN;

	
	  PROC MEANS DATA=&dplocal_lib..&prefix0_dp.sum_vars_by_st_time  NWAY NOPRINT;
	   &class_strata_vars. ;
			VARS   &independent_vars. _SUMWGT_;
	    OUTPUT 	 OUT=&prefix0_dp.sum_vars_by_st(DROP=_TYPE_ _FREQ_)
						   SUM= ;
	  RUN;

	   /* The means by  strata will be used to calculate survival function for each strata */

	  DATA 	&tbl_mean_vars_by_st_out.;
	       SET &prefix0_dp.sum_vars_by_st;
		  ARRAY var_lst{*}  &independent_vars.;
		 
			DO i=1 TO DIM(var_lst);
			  var_lst{i}=var_lst{i}/_SUMWGT_ ;
			END;
			DROP i;
		RUN;

	 %END;	 /* End of do-group for &iter_nb=1*/

 /* Extract matrices S^(l), l=0,1,2 and add variables _d_ (number of ties ) _d_w_ (weighted number of ties)*/

  DATA  &prefix_dp.sscp_S ;
    MERGE    &prefix_dp.sscp_in_srt(DROP=_WEIGHT_)
                    &dplocal_lib..&prefix0_dp.Vector_d(KEEP=&strata_vars. &dependent_vars. _d_ _d_w_)
			;
	   	BY &strata_vars. &dependent_vars.;

	   IF _TYPE_="SSCP";
	   /* Initiate _ties_index_ used by Efron approximation to 1. It will stay this way for Breslow approximation. */
	   _ties_index_=1;
 RUN;

    /* Calculate S^(l) matrices for Effron approximation */

   %IF &ties_sh.=e %THEN	 
               %DO;

		DATA  &prefix_dp.sscp_S_ ;
		  SET  &prefix_dp.sscp_S;
		  DROP _ties_index_;
		RUN;

		%LET _new_names_str=;

	%modify_colnames(tbl_name=&prefix_dp.sscp_S_,var_lst=Intercept &independent_vars.,prefix=,suffix=__s);

	%LET S_names=&_new_names_str.;
	
	%PUT  S_names=&S_names.;

	/* Extract Q^(l) matrices. */

	  DATA  &prefix_dp.sscp_Q ;
        SET &prefix_dp.sscp_in_srt;
	      IF _TYPE_="SSCP_Q";
		  KEEP &strata_vars.  &dependent_vars. 	 _NAME_	 Intercept &independent_vars.;
       RUN;

		%LET _new_names_str=;

	%modify_colnames(tbl_name=&prefix_dp.sscp_Q,var_lst=Intercept &independent_vars.,prefix=,suffix=__q);

	%LET Q_names=&_new_names_str.;
	
	%PUT  Q_names=&Q_names.;

	    /* Merge Breslow S^(l) matrices with  Q^(l) matrices to obtain Effron S^(E,l) matrices */

		DATA  &prefix_dp.sscp_S ;
		  	MERGE  &prefix_dp.sscp_S_ 
			               &prefix_dp.sscp_Q
				;
		  BY &strata_vars.  &dependent_vars. _NAME_	;

		   ARRAY var_lst{*} Intercept &independent_vars.;
		   ARRAY var_s{*} &S_names.;
		   ARRAY var_q{*} &Q_names.;
	
		/* Set _ties_index_ used by Efron approximation and redefine SSCP matrix elements
		    in accordance with this approximation. */

		  DO _ties_index_=1 TO _d_;
					    DO _i_=1 TO DIM(var_lst);
					      var_lst{_i_}=var_s{_i_}  - var_q{_i_} * (_ties_index_-1) /_d_ ;				
					   END;	
			   	OUTPUT;
			END;

		KEEP &strata_vars. &dependent_vars. _NAME_ _ties_index_	Intercept &independent_vars. _d_ _d_w_;
		RUN;

             %END; 	  /* End of do-group for &ties_sh.=e (Efron approx. )*/
   
		   /* Extract from the dataset with matrices S^(l), l=0,1,2 the following  S^(2) matrix
			  , S^(1) vector (LOWCASE(_NAME_)="intercept"). In the latter the column "Intercept"
			   stores scalar  S^(0) (RENAME=(Intercept=_S_) ) */

	  DATA 	&prefix_dp.Matrix_S2(DROP=Intercept 	)
				 &prefix_dp.Vector_S1(DROP= _NAME_   RENAME=(Intercept=_S_))
			   ;
			SET &prefix_dp.sscp_S/*(DROP=_WEIGHT_)*/;

	
			IF LOWCASE(_NAME_)="intercept" THEN 
                                                             DO;
														OUTPUT &prefix_dp.Vector_S1;
															END;
			 ELSE
			  											   DO;
													   OUTPUT &prefix_dp.Matrix_S2;
														   END;
	   RUN;

	   
	   DATA &prefix_dp.Vector_S1a;
		 SET &prefix_dp.Vector_S1(DROP=_S_ _d_ _d_w_ );

		  _FREQ_for_sscp_  =1	;
		  _WEIGHT_for_sscp_=1;

		  /* The transformation below is done to avoid warning message from PROC CORR when _FREQ_=1
			  It does not change resulting sscp matrix because it depends on the product of  f*w
		  */
		   IF   _FREQ_for_sscp_ =1  THEN
							  DO;
			   _FREQ_for_sscp_=2;
			    _WEIGHT_for_sscp_=0.5*_WEIGHT_for_sscp_;
							END;
	   RUN;
 
			%LET _new_names_str=;

	%modify_colnames(tbl_name=&prefix_dp.Vector_S1,var_lst= &independent_vars.,prefix=,suffix=_s1);

	%LET S1_names=&_new_names_str.;
	
	%PUT  S1_names=&S1_names.;

	   	/* Create macro variable with the list of renamed names. It will be used in the step below 
	for variables coming from &dplocal_lib..&prefix0.Vector_d  */

	%modify_colnames(var_lst= &independent_vars.,prefix=,suffix=_d1);

	%LET d1_names=&_new_names_str.;
	
	%PUT  d1_names=&d1_names.;

	/* Calculate contribution to gradient on &strata_vars. &dependent_vars.  level */

	    DATA   &dplocal_lib..&prefix_dp.grad_t_lev;
		 MERGE    &prefix_dp.Vector_S1
						  &dplocal_lib..&prefix0_dp.Vector_d(KEEP=&strata_vars. &dependent_vars. &d1_names. ) 
						  &dplocal_lib..&prefix0_dp.event_set_csum(KEEP=&strata_vars. &dependent_vars. _S0B_ ) 
                    ;

		    BY &strata_vars. &dependent_vars. ;
			

		   ARRAY var_lst{*}  &independent_vars.;
		   ARRAY var_d1{*} &d1_names.;
		   ARRAY var_s1{*} &S1_names.;
		 
		   
			   DO i=1 TO DIM(var_lst);
			       /* For Breslow approximation*/
			      var_lst{i}=var_d1{i} - _d_w_ * var_s1{i}/_S_ ;
					  /* For Efron approximation */
				   IF "&ties_sh."="e" THEN  var_lst{i}=var_lst{i}/_d_; 						 			     
			END;

			 IF "&ties_sh."="e" THEN _S0_=_S0B_ -_d_w_ * (_ties_index_-1) /_d_  ;
			 ELSE  _S0_=_S0B_;

		   KEEP &strata_vars. &dependent_vars. &independent_vars. _ties_index_ _d_  _d_w_ _S_ _S0_ &d1_names. &S1_names. ;
	  RUN;

	  	/* Calculate gradient by summing up contribution from each stratum (except dp_cd if part of strata)
	        and event time. If model is stratified by dp_cd then this calculations are done at each DP
	        and we obtain contribution to gradient from a given DP.
	  */

	PROC MEANS DATA=&dplocal_lib..&prefix_dp.grad_t_lev NWAY NOPRINT;
	     VARS &independent_vars.;
	    OUTPUT 	 OUT=&tbl_Gradient_out.(DROP=_TYPE_ _FREQ_ )
						   SUM=;
	  RUN;
	QUIT;

	/* Calculate contributions to log-likelihood ( _L__Sum) , log-likelihood without covariates (_L0__Sum)
	  and also contributions to number of ties and sum of covariates.*/

	DATA &prefix_dp._L__Sum;
	   SET &dplocal_lib..&prefix_dp.grad_t_lev(KEEP=&strata_vars. &dependent_vars. _d_ _d_w_ _S_ _S0_ &d1_names. )
           ;
		 IF _N_=1 THEN SET  &tbl_model_coeff_in.(KEEP=&independent_vars.) ;
											
		  ARRAY var_d1{*} &d1_names.;
		 ARRAY betas{*}    &independent_vars.;

			_Xbeta_d1_=0;

				 DO i=1 TO DIM(var_d1);
		_Xbeta_d1_=	_Xbeta_d1_ + betas{i} * var_d1{i};
				 END;

		    /* For Breslow approximation*/
			     _L__Sum=_Xbeta_d1_	-   _d_w_  * log( _S_);
		         _L0__Sum=	-  _d_w_  * log( _S0_);
					  /* For Efron approximation */
				   IF "&ties_sh."="e" THEN  
				                      DO;
                          _L__Sum=_L__Sum/_d_; 
						 _L0__Sum=_L0__Sum/_d_; 
									 END;

		 _LL__Sum=_L__Sum; /* For consistency with genmod */
		 KEEP &strata_vars.  &dependent_vars. _L__Sum _L0__Sum  _LL__Sum;
	RUN;
	QUIT;
		  
	PROC MEANS DATA=&prefix_dp._L__Sum NWAY NOPRINT;
	VARS _L__Sum _L0__Sum  _LL__Sum;
	    OUTPUT 	 OUT=&tbl_resid_sum_out.(DROP=_TYPE_ _FREQ_ )
						   SUM=;
	  RUN;	

	  /* Calculate product of S^(1) * [S^(1)]^T. This is done by calculating 
	     cross-products of columns in the dataset with  S^(1). The calculations
	    are done on the level of  &strata_vars.  &dependent_vars. _ties_index_
	  */
 		
   %calculate_sscp(tbl_name_in=&prefix_dp.Vector_S1a 
                         ,independent_vars=&independent_vars.
						 ,by_vars=&strata_vars.  &dependent_vars. _ties_index_
						 ,freq= _FREQ_for_sscp_ 
                         ,weight=_WEIGHT_for_sscp_
						 ,tbl_name_out=&prefix_dp.Matrix_S1S1_a);

	/* Exclude column  Intercept and row with LOWCASE(_NAME_) NE "intercept
		This come from column of 1	which PROC CORR (used by  %calculate_sscp)
		implicitly adds to the input dataset.
	*/

	  DATA	&prefix_dp.Matrix_S1S1;
		SET &prefix_dp.Matrix_S1S1_a;

		_NAME_=LOWCASE(_NAME_);
		 IF _TYPE_ IN ("SSCP") AND LOWCASE(_NAME_) NE "intercept";

		DROP Intercept  _TYPE_; 
	   RUN;

	  			%LET _new_names_str=;

	%modify_colnames(tbl_name=&prefix_dp.Matrix_S1S1,var_lst= &independent_vars.,prefix=,suffix=_S11);

	%LET S11_names=&_new_names_str.;
	
	%PUT  S11_names=&S11_names.;

   	%LET _new_names_str=;

	%modify_colnames(tbl_name=&prefix_dp.Matrix_S2,var_lst= &independent_vars.,prefix=,suffix=_S2);

	%LET S2_names=&_new_names_str.;
	
	%PUT  S2_names=&S2_names.;

 
	PROC SORT DATA=&prefix_dp.Matrix_S1S1;
	  	BY &strata_vars. &dependent_vars. _ties_index_ _NAME_ ;
	 RUN;

	 PROC SORT DATA=&prefix_dp.Matrix_S2;
	  	BY &strata_vars. &dependent_vars. _ties_index_ _NAME_ ;
	 RUN;
	 
	 /* Merge S^(2) matrix with product S^(1) * [S^(1)]^T by strata, event time and 
	    _ties_index_ (for Effron) and _NAME_ (covariate's name)*/

	   DATA   &prefix_dp.Matrix_S2_S1S1;
		   MERGE &prefix_dp.Matrix_S2
						&prefix_dp.Matrix_S1S1
 				          ;
		BY &strata_vars. &dependent_vars. _ties_index_ _NAME_ ;
	  RUN;

	  	
	  	/* Calculate contribution to Hessian matrix on &strata_vars. &dependent_vars.  level 
	   What we calculate below is usually called information matrix I rather than Hessian matrix
       The latter is differ by sign. To be consistent we use appropriate sign 
	   in the Newton-Raphson equation*/

	   DATA   &prefix_dp.Hessian_t_lev	 ;
		   MERGE    &prefix_dp.Matrix_S2_S1S1
 						   &prefix_dp.Vector_S1(KEEP= &strata_vars. &dependent_vars. _ties_index_ _S_)
 				          ;
		BY &strata_vars. &dependent_vars. _ties_index_;

		   ARRAY var_lst  &independent_vars.;
		   ARRAY var_S2 &S2_names.;
		   ARRAY var_S11 &S11_names.;
		   
			DO i=1 TO DIM(var_lst);
			    /* For Breslow approximation*/
			       var_lst{i}= _d_w_ * (  var_S2{i} - var_S11{i}/_S_)/ _S_;
					  /* For Efron approximation */
				   IF "&ties_sh."="e" THEN  var_lst{i}=var_lst{i}/_d_; 
			END;

		   KEEP &strata_vars. &dependent_vars. _ties_index_ _NAME_ &independent_vars.;
	
	  RUN;

	  /* Calculate Hessian by summing up contribution from each stratum (except dp_cd if part of strata)
	        and event time. When model is stratified by dp_cd then this calculations are done at each DP
	        and we obtain contribution to Hessian from a given DP.
	  */
	  
PROC MEANS DATA=&prefix_dp.Hessian_t_lev NWAY NOPRINT;
  CLASS _NAME_;
	VARS &independent_vars.;
	    OUTPUT 	 OUT=&prefix_dp.Hessian0(DROP=_TYPE_ _FREQ_  )
						   SUM=;
	  RUN;
	QUIT;

		 				   /* Hessian matrix sorted in the order of VAR_NUMB_POS*/
	 PROC SQL;
		   CREATE TABLE  &tbl_Hessian_out.(DROP=VAR_NUMB_POS) AS
	  		  SELECT * 							
                             ,INDEXW(LOWCASE(STRIP("&independent_vars.")),LOWCASE(STRIP(_NAME_)))  AS VAR_NUMB_POS
							FROM &prefix_dp.Hessian0
				ORDER BY  CALCULATED VAR_NUMB_POS
					 ;
		 QUIT;

			/* Calculate  baseline hazard  function for each stratum */

%IF &iter_nb. = %eval(&iter_nb_at_conv. + 1) %THEN
                                            %DO;
	 
   	DATA  &tbl_baseln_hazard.;					
		SET &dplocal_lib..&prefix_dp.grad_t_lev(KEEP=&strata_vars. &dependent_vars. _ties_index_ _d_ _d_w_ _S_ );
		   BY &strata_vars.  &dependent_vars. _ties_index_;

			   RETAIN _Cum_Hazard0_ 0 ;

				IF _d_w_=0              THEN _Hazard_=0;
			    ELSE IF  _S_ =0	THEN _Hazard_=.;
		   	    ELSE 	                _Hazard_=_d_w_ /  _S_	;

				 IF "&ties_sh."="e" THEN _Hazard_=_Hazard_/_d_;

				/*Reset_Cum_Hazard0_  for new Strata  */
				IF &fst_by_strata_var. =1 THEN  _Cum_Hazard0_ =0;

			   _Cum_Hazard0_=SUM(_Cum_Hazard0_,_Hazard_);
			   _LOGSURV0_= -_Cum_Hazard0_;
			 
			   IF LAST.&dependent_vars.;

                 DROP  _Hazard_ _ties_index_;
	RUN;

  %END;	 /* End of do-group for  &iter_nb. = %eval(&iter_nb_at_conv. + 1)  */

  %PUT;
   %PUT -----   End of the macro calc_cox_grad_hessian  iter_nb=&iter_nb. ;
   %PUT;

	 %MEND;


