  %MACRO calc_genmod_vars(tbl_name_in=
												  ,Model_Coeff=
												  ,independent_vars=
												  ,dependent_vars=
					                              ,regr_type_cd=1
												  ,dp_cd=1
												  ,freq=
					                              ,weight=
												  ,tbl_name_out=); 
   
			%LOCAL  weight freq2 weight2;

 /*	 Usage:
	
		 %calc_genmod_vars(tbl_name_in=&reg_ds_dp_in.
									  ,Model_Coeff=infolder.&prefix_prev.Model_Coeff
									  ,independent_vars=&independent_vars.
									  ,dependent_vars=&dependent_vars.
		                              ,regr_type_cd=1
									  ,freq=&freq.
		                              ,weight=&weight.
									  ,tbl_name_out=&prefix_dp.genmod_vars); 
  */

			 /* Add linear predictor to regression dataset using current estimates of \beta */

PROC DATASETS LIB=DATA_IN  NOLIST NOWARN;
   DELETE _REGR_DS_XBETA;
RUN;

PROC SCORE DATA= &tbl_name_in. 
					     SCORE=&Model_Coeff.	  TYPE=PARMS 
						OUT=_regr_ds_xbeta /*(RENAME=(MODEL1=)) */
		;
   VAR &independent_vars.;
run;

 %IF  %LENGTH(&freq.) >0  		%THEN %LET  freq2=&freq.;
 %ELSE 												  		%LET  freq2=1;
  %IF  %LENGTH(&weight.) >0  %THEN 	%LET  weight2=&weight.;
	%ELSE 													%LET  weight2=1;

		   DATA  &tbl_name_out.;
			  SET _regr_ds_xbeta;
			     IF _N_=1 THEN SET &Model_Coeff.(KEEP=_PHI_);

					_pi_=CONSTANT("pi"); 

					regr_type_cd=&regr_type_cd.;
					  dp_cd=&dp_cd.;
					_RESP_=&dependent_vars.;
				    _FREQ_=&freq2.;
				   _WEIGHT_=&weight2.;
				/*	_PHI_=_RMSE_**2;  */

		    ARRAY indep{*}  &independent_vars. ;

 /*Set _XBETA_  to missing when any of &independent_vars is missing. In most cases, PROC SCORE
	does this correctly. However, when of a regression coefficient is 0 PROC SCORE just ignores it 
	and produces non-missing _XBETA_.  The code below corrects this.*/
	
       DO i=1 TO DIM(indep);
			IF indep{i}=. THEN _XBETA_=.;	
	END;

		IF  _RESP_ = .  OR _XBETA_ = .	  THEN  DO; _FREQ_=.; _WEIGHT_=. ; END;

		/*Define very small number. It is used to avoid numerical exception in Log and devision */
					_sml_=1E-12;
			  _RESP1_=MAX(ABS( _RESP_), _sml_); 
			 _1MRESP1_=MAX(ABS( 1-_RESP_) , _sml_); 
	  /* We define _DEV_ as scaled deviance (devided by _PHI_)*/

	  /*In the SAS documentation for GENMOD expressions for _LL_ are always for _PHI_=1
			while expression _L_ is for general case. The actual expressions alwayse have _PHI_ in 
	    	the combination with (_WEIGHT_/_PHI_) in both _LL_ and _L_. For logistic they are identical 
			and for normal and Poisson they differ due to the contribution from normalization constant.  
	*/

  IF  _WEIGHT_ > 0 THEN 
                              DO;
					 										   /* Linear regression */
				   IF regr_type_cd=1 THEN
				           DO;
			   	 _MEAN_=_XBETA_;
				_MEAN_DERV_=1;
				  _L_=-0.5*( _WEIGHT_*(_RESP_ - _MEAN_)**2/_PHI_ +  log(2*_pi_*_PHI_/_WEIGHT_) );
				_LL_=_L_;
				_DEV_= _WEIGHT_*(_RESP_ - _MEAN_)**2/_PHI_ ;
							END;
		
					                                                  /* Logistic regression */
				ELSE IF regr_type_cd=2 THEN
				           DO;
						     /* Defensive coding */
			     IF _XBETA_ >30 THEN _XBETA_ =30;

			   	 _MEAN_=EXP( _XBETA_)/(1+EXP( _XBETA_));
				_MEAN_DERV_=_MEAN_ * (1-_MEAN_);
				 _L_=_WEIGHT_* (_RESP_ *  _XBETA_ -LOG(1+EXP( _XBETA_) ) )/ _PHI_ ;	 
				_LL_=_WEIGHT_* (_RESP_ *  _XBETA_ -LOG(1+EXP( _XBETA_) ) )/ _PHI_;	
_DEV_= -2*(_L_  - _WEIGHT_*(  _RESP1_ * LOG(_RESP1_ )+_1MRESP1_ * LOG(_1MRESP1_) )/ _PHI_  );
 
							END;
																   /* Poisson regression  */
				  ELSE IF regr_type_cd=3 THEN
				           DO;
				  /* Defensive coding */
			  IF _XBETA_ >30 THEN _XBETA_ =30;

			   	 _MEAN_=EXP( _XBETA_);
				_MEAN_DERV_=_MEAN_;	
				 _L_=_WEIGHT_*( _RESP_ *  _XBETA_ - _MEAN_ - LFACT(INT(_RESP_)) )/ _PHI_;
				_LL_= _WEIGHT_*( _RESP_ *  _XBETA_ - _MEAN_ - LFACT(INT(_RESP_)) )/ _PHI_;
				 _DEV_=2* _WEIGHT_*(_RESP1_ * (LOG(_RESP1_) - _XBETA_)  +_RESP_ - _MEAN_ )/ _PHI_;
							END;

			/* Canonical link implies _VARIANCE_=_MEAN_DERV_;	*/
				  _MEAN_DERV_=MAX(_MEAN_DERV_, _sml_);
				_VARIANCE_=_MEAN_DERV_;	

				_ADJ_VARIANCE_=_VARIANCE_ * _PHI_/_WEIGHT_;

				 _WEIGHT_for_SSCP_=_WEIGHT_*(_MEAN_DERV_)**2/_VARIANCE_;

		/* Redefine &dependent_vars. for SSCP GENMOD  calculation. Original is kept in variable _RESP_ */
				 &dependent_vars.=(_RESP_ - _MEAN_+_MEAN_DERV_*_XBETA_)/_MEAN_DERV_ ;	
	
				_RESP_W_=_WEIGHT_ * _RESP_;
			    _MEAN_W_ =_WEIGHT_ * _MEAN_;
				_RESIDUAL_=_RESP_ - _MEAN_;
			    _RESIDUAL_SQ_=_RESIDUAL_ **2;
		 /* Gradient Residual. For canonical link and _WEIGHT_=1 it is the same as _RESIDUAL_ */
			    _GRESIDUAL_=(_RESIDUAL_)*_MEAN_DERV_/_ADJ_VARIANCE_;
    /* Pearson Residual. If model assumptions about variance are correct then  variance of _PRESIDUAL_ is 1*/
			   _PRESIDUAL_=_RESIDUAL_/SQRT(_ADJ_VARIANCE_);
			    _PRESIDUAL_SQ_=_PRESIDUAL_**2;
				_DRESIDUAL_=SQRT(ABS(_DEV_)) *SIGN(_RESIDUAL_);
			    _DRESIDUAL_SQ_=_DRESIDUAL_**2;

					/* Matrix I1 will be used for Robust sandwich variance estimator(RSVE) 
				     We do not use scaling factor in the expression for _WEIGHT_for_I1_MATR_ because 
				      we  do not use it in the _WEIGHT_for_SSCP_ either.The importent thing is that
					 the final expression for RSVE=InvXPX*I1_Matrix*InvXPX does not depend on  _PHI_.
 					 The only thing that depend on _PHI_ is the model based covariance matrix of 
					  estimates and correspondinf StdErr. They are adjusted in cental module at the
 					  final step.
		 	*/
		   	_WEIGHT_for_I1_MATR_=_WEIGHT_ * _WEIGHT_for_SSCP_ * _RESIDUAL_**2 /_VARIANCE_;
				          END;
	 ELSE 		&dependent_vars.=.;  /* Redefine &dependent_vars. for SSCP calculation. */

			 DROP i  _pi_ _sml_ regr_type_cd;
		   RUN;

  %MEND;   
