 /*HEADER------------------------------------------------------------------------
|
| PROGRAM:  dp_main (macro)
||
| PURPOSE:  The main macro which does processing necesary for distributed regression at a DP
			           site. In the production enviroment it has to be called from the master program which 
                       defines Sentinel site specific parameters and libraries.

                       This macro calls other macros which perform specific to a type of 
                       regression (linear, logistic, etc.) or analysis (e.g. residual stats).

					   Most of the parameters for this macro come from the central site via 
                      parameter dataset: vars_nm_value_pairs

 					   The macro has one parameter dp_cd wich identifies numeric code for a DP.
					  The mapping of these codes for Sentinel DPs is defined in the table: map_dpid_siteid_to_dp_cd.
					  The table is located in infolder library and is distributed with the package.
					  The user does not need to specify the code explicitly. The parameter 	dp_cd
					  is derived in the master wrapper program using standard Sentinel codes for a DP. 

	First version:  04/07/2016, Yury Vilk 
	Ver 2. After 20170130 (modification for job_done.ok job_fail) . Y.V.
+------------------------------------------------------------------------HEADER*/

%MACRO	 dp_main(dp_cd=);
		   	
		 %PUT;
		 %PUT -----   Start of the macro dp_main for DP dp_cd: &dp_cd.;
		 %PUT;

			  %LOCAL runid reg_ds_dp_in reg_ds_file_name_sh  
							 m_vars_list_dp  _sscp_input_ds _sscp_output_ds_sh  weight_sscp
							prefix0 prefix0_dp prefix_dp prefix_prev tbl_list_to_receive tbl_list_to_transf  
                            last_iter_in iter_nb step_nb reg_conv_in reg_conv_msg last_runid_in regr_type_cd
 							 addn_tbl_to_receive addn_tbl_to_transf
                            step_return_cd step_return_msg runid_return_cd freq weight
						    pos_tmp wait_time_min wait_time_max
						    rc_msoc rc_dplocal rc_infolder rc_sasmacr
							nobs_vars_nm_value_pairs log_file_sh_nm
						   end_job_dp_in  job_end_file_nm 
						   runid_return_dp_msg	censoring_var
                             ;
		  /* Initialize variables. This will be overwritten below using the dataset 
	                  VARS_NM_VALUE_PAIRS coming from central.               
	             */
	%let RunID=;
    %let iter_nb=0;
	%let step_nb=1;
	%let  step_return_cd=0;
 	%let  step_return_msg=PASS;
    %let reg_conv_in=0; 
    %let reg_conv_msg=;
	%LET last_iter_in=0;
    %let last_runid_in=1;
	%let regr_type_cd=0;
	%LET end_job_dp_in=0;
	%LET censoring_var=;

	  %LET m_vars_list_dp=;

	  %LET prefix0=&runid._;
	  %LET prefix=&prefix0.i&iter_nb._;
	  
	   /* Initialize local  &wait_time_min   and &wait_time_max	and  prefix_prev
	   This will be set later based on &INFOLDER_LIB..VARS_NM_VALUE_PAIRS */

	    %LET wait_time_min=3;
    	%LET wait_time_max=20000;
		%LET prefix_prev=&prefix.;

		/* Initialize local &runid_return_dp_cd */
	    %LET runid_return_dp_cd=0; 

      %LET rc_msoc=%SYSFUNC(FILEEXIST("&msoc_dir."));
	   %LET rc_dplocal=%SYSFUNC(FILEEXIST("&dplocal_dir."));
	   %LET rc_infolder=%SYSFUNC(FILEEXIST("&infolder_dir."));
	   %LET rc_sasmacr=%SYSFUNC(FILEEXIST("&sasmacr."));

	%IF       &SYSCC. > 4	
            OR &rc_msoc.     NE 1 
			OR &rc_dplocal.  NE 1 			
			OR &rc_infolder.  NE 1
			OR &rc_sasmacr. NE 1
      %THEN 
							%DO;
   %LET runid_return_dp_cd=2100;
   	%LET runid_return_dp_msg=%SYSFUNC(COMPBL(%STR(Error:  Main DP program at the site dp_cd: &dp_cd. will stop executing. 
              The most likely source of errors are incorrect references to: folders, file names or SAS libraries:
			  rc_msoc=&rc_msoc. rc_dploca=&rc_dplocal. rc_infolder=&rc_infolder. rc_sasmacr=&rc_sasmacr.
                                                                         )));
	   	%PUT;
   		%PUT runid_return_dp_msg=&runid_return_dp_msg.;
   	    %PUT;
				%GOTO endmac_fatal_err;
			  				%END;
	
   	%delete_ext_file(file_name=&msoc_dir.job_started.ok, warning_flag=0);

		%IF %LENGTH(&step_return_cd.)=0 OR &step_return_cd.=0 %THEN 
					%DO;
					  %create_empty_file(file_nm=&msoc_dir.job_started.ok);
	%IF &m_return_cd. >0 %THEN  
										%DO;
			/* If there is no write permissions to the folder &msoc_dir. */
	%LET runid_return_dp_cd=2099;
   	%LET runid_return_dp_msg=%STR(Error:  Main DP program at the site dp_cd= &dp_cd. will stop executing. Can not write to the folder &msoc_dir..);
	   	%PUT;
   		%PUT runid_return_dp_msg=&runid_return_dp_msg.;
   	    %PUT;

						%GOTO endmac_fatal_err;
					  				    %END;
					   %END;

	/* Initial cleanup to make sure that files from previous run/iteration will not be used*/
	  %delete_ext_file(file_name=&msoc_dir.job_done.ok, warning_flag=0);
       %delete_ext_file(file_name=&msoc_dir.job_fail.ok, warning_flag=0);

	   %delete_sas_tables(tbl_name_sh_list=_i0_: );
	     %delete_sas_tables(lib=&msoc_lib.,tbl_name_sh_list=_i0_: );

	   %DO %UNTIL(/*&last_iter_in.=1 OR*/ &exit_after_each_iter.=1 OR &end_job_dp_in. =1);

				    %LET  step_return_cd=0;
 	                %LET  step_return_msg=PASS;

					   /* Vars in file_watcher macro must be local to DP. */

	%file_watcher(file_name=&infolder_dir.files_done.ok
		 					 ,wait_time_min=&wait_time_min.
							 ,wait_time_max=&wait_time_max.
							);

	 %IF &_fileexist_flag.=0 %THEN  
										%DO;
			/* If wait time for file files_done.ok exceeded  wait_time_max*/
	%LET runid_return_dp_cd=2098;
   	%LET runid_return_dp_msg=%STR(Error: Main DP program at the site dp_cd= &dp_cd.. Wait time for file files_done.ok exceeded  wait_time_max=&wait_time_max. sec .);
	   	%PUT;
   		%PUT runid_return_dp_msg=&runid_return_dp_msg.;
   	    %PUT;
						%GOTO endmac_fatal_err;
					  				    %END;
  
	 /* Defensive coding. Should be removed by DP's PMN at the end of each iteration */

	   %delete_ext_file(file_name=&msoc_dir.files_done.ok, warning_flag=0);
	   %delete_ext_file(file_name=&msoc_dir.file_list.csv, warning_flag=0);
	
		%isdata(dataset= &INFOLDER_LIB..VARS_NM_VALUE_PAIRS);

			%LET nobs_vars_nm_value_pairs=&nobs.;

		%IF &nobs. >0  %THEN
		                                  %DO;
		 PROC SQL NOPRINT;
		   SELECT STRIP(M_VAR_NM)   INTO  :m_vars_list_dp	 SEPARATED BY " "
			 FROM  &INFOLDER_LIB..VARS_NM_VALUE_PAIRS
				 ORDER BY  M_VAR_ID
				 ;
		  QUIT;

		   %LOCAL &m_vars_list_dp. ; 

		  %PUT 	m_vars_list_dp=&m_vars_list_dp.;
		 
			   DATA _NULL_;
				  SET &INFOLDER_LIB..vars_nm_value_pairs;
				  CALL SYMPUT(M_VAR_NM,STRIP(M_VAR_VALUE));
					PUT   M_VAR_NM " = "  M_VAR_VALUE;
			   RUN;

	    %PUT;
		 %PUT --- Read input parms table VARS_NM_VALUE_PAIRS for  DP: dp_cd=&dp_cd. iter_nb=&iter_nb. step_nb=&step_nb. ---;
		 %PUT     test_env_cd=&test_env_cd.;
		 %PUT;
											 %END; 
	  %ELSE    
												%DO;
					   %LET step_return_cd=20;
					 %LET  step_return_msg=%STR(Error: parameter dataset vars_nm_value_pairs. does not exist dp_cd=&dp_cd..) ;
					  %PUT;
					  %PUT  step_return_msg=&step_return_msg.;
					  %PUT;				
				 									%END;

	%PUT  Read input table. Processing dp_cd=&dp_cd. iter_nb=&iter_nb. step_nb=&step_nb.;

%IF               %SYMEXIST(min_count_per_grp) =0  %THEN %LET min_count_per_grp=&min_count_per_grp_glob.;
 %ELSE %IF %LENGTH(&min_count_per_grp.)=0 %THEN %LET min_count_per_grp=&min_count_per_grp_glob.;

			 %PUT min_count_per_grp=&min_count_per_grp. ;

		    /*For initial testing */
			/*  %LET DP_CD=1;	*/

	  %LET prefix0=&runid._;
	  %LET prefix=&prefix0.i&iter_nb._;		

	   					/* For Prod and &test_env_cd.=0, 2*/
			%LET  prefix0_dp=&runid._;
			%LET  prefix_dp=&prefix. ;
			%LET reg_ds_dp_in=&data_in_lib..&reg_ds_in.;

	 	%IF  &test_env_cd.=1 %THEN 
		                   %DO;
			/*For test. Central dir structure */
			 %LET  prefix0_dp=&runid._dp&dp_cd._;
			 %LET  prefix_dp=&prefix.dp&dp_cd._ ;
		    %LET reg_ds_dp_in=&data_in_lib..&reg_ds_in._&dp_cd.;
			 %LET exit_after_each_iter=1;	
						   %END;

			%LET pos_tmp=%EVAL(%INDEX(&reg_ds_dp_in.,.) +1);
           %LET reg_ds_file_name_sh=%LOWCASE(%SUBSTR(&reg_ds_dp_in,&pos_tmp.).sas7bdat);

        %PUT reg_ds_file_name_sh=&reg_ds_file_name_sh.;


/* Initial cleanup to make sure that files from previous run/iteration will not be used. &RunID level*/

 %IF &iter_nb.=1 %THEN
  										%DO;
		 %delete_sas_tables(tbl_name_sh_list=&prefix0_dp.: );
	     %delete_sas_tables(lib=&msoc_lib.,tbl_name_sh_list=&prefix0_dp.: );
	     %delete_sas_tables(lib=&dplocal_lib.,tbl_name_sh_list=&prefix0_dp.: );

									   %END; /* End of do-group &iter_nb.=1*/

 /* Initial cleanup to make sure that files from previous run will not be used(iteration level)*/

       %delete_ext_file(file_name=&msoc_dir.file_list.csv, warning_flag=0);
        %delete_ext_file(file_name=&msoc_dir.files_done.ok, warning_flag=0);

	    %delete_sas_tables(tbl_name_sh_list= &prefix_dp.:);
	     %delete_sas_tables(lib=&msoc_lib.,tbl_name_sh_list= &prefix_dp.:);
	     %delete_sas_tables(lib=&dplocal_lib.,tbl_name_sh_list= &prefix_dp.:);	

  	%IF &nobs_vars_nm_value_pairs. >0  %THEN
		                                  %DO;
 		DATA &dplocal_lib..&prefix_dp.VARS_NM_VALUE_PAIRS;
	       SET &INFOLDER_LIB..VARS_NM_VALUE_PAIRS; 
	   RUN;
										   %END; 

	         /* Initialize tracking table for a given itteration &iter and &dp_id */

	     %initialize_tracking_tbl(dp_cd=&dp_cd.,tbl_name_out=&msoc_lib..&prefix.tr_tb_updt);

		 
		  %LET _sscp_output_ds_sh=&prefix.sscp.sas7bdat;
		 %LET _sscp_output_ds=&msoc_lib..&prefix.sscp;	 
	
  		%LET addn_tbl_to_receive=;
        %LET addn_tbl_to_transf=;

 /* For Cox regression */
	%IF &regr_type_cd. =10   %THEN
				        %DO;
		  	  	
		   %IF       (&exist_init_events_time_set. =1 AND &iter_nb.= 1)
	 			  OR (&exist_init_events_time_set. =0 AND &iter_nb.= 2 AND  &pos_dp_cd_in_strata_vars. =0)
			%THEN 	%LET addn_tbl_to_receive=&tbl_events_time_set..sas7bdat;

  /* Note that the condinition &iter_nb. > %eval(&iter_nb_at_conv. + 1) can only be satisfied
      if &pos_dp_cd_in_strata_vars. =0 and thus &numb_iter_after_conv=2 */

		%IF	&iter_nb. <= %eval(&iter_nb_at_conv. + 1) 
	     %THEN 	 
		                               %DO;
        %LET addn_tbl_to_receive=%SYSFUNC(COMPBL(&addn_tbl_to_receive. 
                                                                         &prefix_prev.model_coeff.sas7bdat));

           %IF &iter_nb. = 1  %THEN %LET addn_tbl_to_transf=&prefix0.total_cnts.sas7bdat;
			 %LET addn_tbl_to_transf=%SYSFUNC(COMPBL(&addn_tbl_to_transf. &prefix.sscp.sas7bdat));
					               
		   %IF &pos_dp_cd_in_strata_vars. >0  %THEN
														%DO;

  %LET addn_tbl_to_transf=%SYSFUNC(COMPBL(&addn_tbl_to_transf. &prefix.resid_sum.sas7bdat));

 %IF &iter_nb. = 1  
 %THEN %LET addn_tbl_to_transf=%SYSFUNC(COMPBL(&addn_tbl_to_transf. &prefix0.mean_vars_by_st.sas7bdat));

%IF &iter_nb. = %eval(&iter_nb_at_conv. + 1) 
  %THEN	%LET addn_tbl_to_transf=%SYSFUNC(COMPBL(&addn_tbl_to_transf. &prefix0.baseln_hazard.sas7bdat 
                                                                                                 &prefix0.resid_sum_by_pct.sas7bdat));

														  %END;	 /* End of do-group  &pos_dp_cd_in_strata_vars. >0*/

										  %END;	 /* End of do-group  &iter_nb. <= %eval(&iter_nb_at_conv. + 1) */
		 %ELSE 	  
								       %DO;
          %LET addn_tbl_to_receive=&addn_tbl_to_receive. &prefix0.baseln_hazard.sas7bdat;
          %LET addn_tbl_to_transf=%SYSFUNC(COMPBL(&prefix0.resid_sum_by_pct.sas7bdat));
									   %END;  /* End of do-group &iter_nb. > %eval(&iter_nb_at_conv. + 1). Only possible when &pos_dp_cd_in_strata_vars. =0 */

						 %END;  /* End of do-group for &regr_type_cd. =10*/
 
						 /* Generalized Linear Model regression */
	   %ELSE 
			 %DO;
				    %IF    &regr_type_cd. =1  %THEN   
	                                        %DO;
             %LET _sscp_input_ds=&reg_ds_dp_in.;
			 %LET weight_sscp=%STR(&weight.);

	%IF &iter_nb. = 1  %THEN %LET addn_tbl_to_transf=&prefix.sscp.sas7bdat ;
	%ELSE 	 
                                                               %DO;
	                      %LET addn_tbl_to_receive=&prefix_prev.model_coeff.sas7bdat;
                          %LET addn_tbl_to_transf=&prefix.resid_sum.sas7bdat ;
	         		                                             %END; /* End of do-group &iter_nb. = 1  */
						                     %END;	/* End of do-group &regr_type_cd. =1   */
	 %ELSE 	    
	 										 %DO;
	   %LET _sscp_input_ds=&prefix_dp.reg_vars;
		%LET weight_sscp=_WEIGHT_for_sscp_;

		 %LET addn_tbl_to_receive=&prefix_prev.model_coeff.sas7bdat;
		%LET addn_tbl_to_transf=%SYSFUNC(COMPBL(&prefix.resid_sum.sas7bdat &prefix.sscp.sas7bdat)) ;
									      		%END;  /* End of do-group &regr_type_cd.  NOT IN (1 10) */

		%IF &iter_nb. = %eval(&iter_nb_at_conv. + 1) %THEN 
                     									%DO;
%LET addn_tbl_to_transf= %SYSFUNC(COMPBL(&addn_tbl_to_transf. &prefix0.sw_matrix.sas7bdat
                                        &prefix0.resid_sum_by_pct.sas7bdat	&prefix0.resid_sum_by_pct2.sas7bdat));
                      								     %END; 

			            %END;  	/* End of do-group &regr_type_cd.  NE 10 (GenMod)*/

 	%LET tbl_list_to_receive=%SYSFUNC(COMPBL(vars_nm_value_pairs.sas7bdat  &addn_tbl_to_receive. ));

	%LET tbl_list_to_transf=%SYSFUNC(COMPBL(&addn_tbl_to_transf.  &prefix.tr_tb_updt.sas7bdat
                                                                                  &prefix.tr_tb_fl_updt.sas7bdat));

%output_file_list(tbl_list_to_transf=
						    ,tbl_list_not_transf=&tbl_list_to_receive.
						    ,file_list_txt_out=
						    ,file_list_tbl_out=&prefix_dp.file_list_in
							,send_received_cd=2
							,dir_nm=&infolder_dir.);

  %find_files_mod_dtm( tbl_name_in=&prefix_dp.file_list_in , tbl_name_out=&prefix_dp.file_list_in_dtm);
	
      /* No parameter dataset vars_nm_value_pairs received from central */
	%IF &step_return_cd. > 0 %THEN
	             %DO;
  %PUT;
  %PUT  step_return_msg=&step_return_msg.;
  %PUT;
%GOTO endmac; 
				 %END;
 
      /*  &runid_return_cd  comes from central site.*/ 
   	%IF &runid_return_cd. > 0 %THEN
	             %DO;
   %LET step_return_cd=10;
 %LET  step_return_msg=Error: Stop command from central site. Run with RunID=&RunID. will terminate. ;
  %PUT;
  %PUT  step_return_msg=&step_return_msg.;
  %PUT;
%GOTO endmac; 
				 %END;

    	    %IF %LENGTH(&missing_or_0_obs_files.) >0 %THEN
                  %DO;
	   %LET  step_return_cd=1;	/* Missing input files*/
	  %LET  step_return_msg=%STR(Error: Step fail dp_cd=&dp_cd. iter_nb=&iter_nb. step_nb=&step_nb..Requiered files were not found or had 0 obs: &missing_or_0_obs_files.);
	   %PUT;
       %PUT  step_return_msg=&step_return_msg.;
      %PUT;
	     %GOTO endmac;  
				  %END;	

   
	  %isdata(dataset=&reg_ds_dp_in.);
	   %IF &nobs.=0 %THEN
	                        %DO;
 	   %LET step_return_cd=1100;
	   %LET runid_return_dp_cd=&step_return_cd.;
	   %LET  step_return_msg=Fatal Error: Input regression dataset &reg_ds_dp_in. does not exist or has zero observation ;
		 %PUT;
       %PUT  step_return_msg=&step_return_msg.;
        %PUT;
						%GOTO endmac;
						   %END;

%check_var_list_in_tbl(tbl_name_in=&reg_ds_dp_in.
                                       , num_vars=&dependent_vars. &independent_vars. &freq. &weight. &censoring_var.) ;

%IF %LENGTH(&missing_vars.) >0 %THEN
		                   %DO;			  
	   %LET step_return_cd=1099;
	    %LET runid_return_dp_cd=&step_return_cd.;
	   %LET  step_return_msg=%STR(Fatal Error: Required variables not found in input dataset &reg_ds_dp_in.: &missing_vars.) ;
	  %PUT;
       %PUT  step_return_msg=&step_return_msg.;
       %PUT;
				      %GOTO endmac;
						   %END;

	%IF &step_return_cd.=0 AND  &SYSCC. > 4  %THEN	 
 				 %DO;
 %LET  step_return_cd=100;
 %LET  step_return_msg=%STR(Error: Step fail dp_cd=&dp_cd. iter_nb=&iter_nb. step_nb=&step_nb.. See log for details);
   %PUT;
   %PUT  step_return_msg=&step_return_msg.;
   %PUT;
		  %GOTO endmac;  
				%END;

%IF &regr_type_cd. = 10 %THEN
	 %DO;     
			  %IF &iter_nb. <= %eval(&iter_nb_at_conv. + 1) %THEN
				            %DO; 
				     %calc_cox_iter_stats_dp;
						  %END; /* End of do-group for &iter_nb. LE %eval(&iter_nb_at_conv. + 1)*/

			  %IF &iter_nb. = %eval(&iter_nb_at_conv. + &numb_iter_after_conv.) %THEN
				              %DO;
				%calc_residual_stats_cox_dp;
							%END;

	 %END;	/* End of do-group for &regr_type_cd. = 10 (Cox model) */

 %ELSE 
 		%DO;     
	               %calc_genmod_iter_stats_dp;

			    %IF &last_iter_in. =1 %THEN
	                                %DO;	
		              %calc_residual_stats_genmod_dp;
								   %END;  /* &last_iter_in. =1*/

            %END; /* End of do-group for &regr_type_cd. NE 10 ( generalized linear model )*/

   %endmac:

     %IF &runid_return_dp_cd.=0 AND &step_return_cd. >0  %THEN
                   %DO;
		  %LET runid_return_dp_cd=&step_return_cd.;
		 %LET runid_return_dp_msg=&step_return_msg.;
		               %END;

 %PUT Update time stamp on &msoc_lib..&prefix.tr_tb_updt  ;;
 %PUT and intialize &msoc_lib..&prefix.tr_tb_fl_updt and &msoc_dir.file_list.csv ;
/*	The time stamps will be recorded below by  macro %find_files_mod_dtm.
   The actual update of these tables will occur a few steps below
*/
		DATA &msoc_lib..&prefix.tr_tb_updt
				 &msoc_lib..&prefix.tr_tb_fl_updt
                 ;
		    SET &msoc_lib..&prefix.tr_tb_updt;
	     RUN; 

   %create_empty_file(file_nm=&msoc_dir.file_list.csv);
   
%export_sas_to_dlm_txt(tbl_name_in=&msoc_lib..&prefix.tr_tb_updt
                                          ,file_name_out=&msoc_dir.&prefix.tr_tb_updt.csv);

 	  %LET log_file_sh_nm=;

     %IF %SYSFUNC(FILEEXIST("&msoc_dir.&MSReqID..log"))=1 
      %THEN %LET log_file_sh_nm=&MSReqID..log;

	%LET tbl_list_to_transf=&tbl_list_to_transf.  &log_file_sh_nm.;
 
 	%PUT log_file_sh_nm=&log_file_sh_nm.;
    %PUT tbl_list_to_transf=&tbl_list_to_transf.;

%output_file_list(tbl_list_to_transf=&tbl_list_to_transf.
						    ,tbl_list_not_transf=file_list.csv &prefix.tr_tb_updt.csv
						    ,file_list_tbl_out=&prefix_dp.file_list_out
							,dir_nm=&msoc_dir.);

%find_files_mod_dtm( tbl_name_in=&prefix_dp.file_list_out , tbl_name_out=&prefix_dp.file_list_out_dtm);

                /*2018020; Added small delay to avoid file lock condition when deleting file   */

        DATA  _NULL_;	
		   CALL SLEEP(3,1);	
		 RUN;

			  /* For production enviroment version.  */
	 %IF &test_env_cd.=0 /*AND  &step_return_cd. < 1000*/ %THEN 
         			%DO;
	%delete_ext_file(file_name=&infolder_dir.files_done.ok, warning_flag=0);
	%delete_sas_tables(lib=&INFOLDER_LIB., tbl_name_sh_list=vars_nm_value_pairs);
		 			%END;

	/* Set non-zero step_return_cd  unless it was already set earlier in the program. */
 
 %IF &step_return_cd.=0 AND  %LENGTH(&missing_or_0_obs_files.) >0 %THEN 
											  %DO;
  %LET  step_return_cd=3;	/* Missing output files*/
  %LET  step_return_msg=%STR(ERROR: Step fail dp_cd=&dp_cd. iter_nb=&iter_nb. step_nb=&step_nb..Requiered files were not created or had 0 obs: &missing_or_0_obs_files.);
   %PUT;
   %PUT  step_return_msg=&step_return_msg.;
   %PUT;
											 %END;
 %ELSE %IF &step_return_cd.=0 AND  &SYSCC. > 4  %THEN	 
 											 	  %DO;
 %LET  step_return_cd=100;
 %LET  step_return_msg=%STR(ERROR: Step fail dp_cd=&dp_cd. iter_nb=&iter_nb. step_nb=&step_nb.. See log for details);
  %PUT;
  %PUT  step_return_msg=&step_return_msg.;
  %PUT;
											    %END;

 /*	 *20170131. Change SAS at DP will exit at the next iteration after central site will set end_job_dp_in=1;
 	%IF &step_return_cd. >0  %THEN           
		                          %DO;					  
					  %LET exit_after_each_iter=1;
					  %LET end_job_dp_in=1;
								  %END;
   */
	    DATA &msoc_lib..&prefix.tr_tb_updt;
		    SET &msoc_lib..&prefix.tr_tb_updt;
			  END_DTM=DATETIME();
			  STEP_RETURN_CD=&step_return_cd.;
		      STEP_RETURN_MSG="&step_return_msg.";
	       RUN; 

%export_sas_to_dlm_txt(tbl_name_in=&msoc_lib..&prefix.tr_tb_updt
                                          ,file_name_out=&msoc_dir.&prefix.tr_tb_updt.csv);


	/* The table &prefix.tr_tb_fl_updt have info about itself. For this table we will  set file_exist_flag
	  and file_mod_dtm a little bit before it is actually created or modified in the next steps. */

	DATA &msoc_lib..&prefix.file_list_dtm;
	  SET &prefix_dp.file_list_in_dtm &prefix_dp.file_list_out_dtm;
	RUN;

  %create_tracking_tbl_fl_lev_updt(	updt_tbl_name_in=&msoc_lib..&prefix.tr_tb_updt
										                    ,file_list_dtm_in=&msoc_lib..&prefix.file_list_dtm
															,updt_tbl_name_fl_lev_out=&msoc_lib..&prefix.tr_tb_fl_updt
															);

 %LET job_end_file_nm=;

	DATA &prefix.file_list_out_for_csv;
	   SET &msoc_lib..&prefix.file_list_dtm;
	     IF         send_received_cd=1 
             AND file_exist_flag=1
              ;

		IF STRIP(LOWCASE(file_nm)) = "file_list.csv"  THEN transfer_to_site_in=10;
			
			   /* Do not transfer log file until the very end */
		IF   file_nm ="&log_file_sh_nm." AND &end_job_dp_in.=0 THEN DELETE;
	
        FORMAT transfer_to_site_in 4.;	
	   KEEP file_nm transfer_to_site_in;
	RUN;


 %IF &end_job_dp_in.=1 %THEN
                     %DO;
	%IF &step_return_cd.=0 %THEN 
                                       %DO;
	%LET job_end_file_nm=job_done.ok;
    %create_empty_file(file_nm=&msoc_dir.job_done.ok);
	                                     %END;
%ELSE											
										%DO;
	%LET job_end_file_nm=job_fail.ok;
	%create_empty_file(file_nm=&msoc_dir.job_fail.ok);
						                  %END;
					 %END;

	DATA   &msoc_lib..&prefix.file_list_out_for_csv;
	 SET &prefix.file_list_out_for_csv END=last;
	    OUTPUT;
		 IF last  THEN
		   DO;
			 IF "&job_end_file_nm." NE " "	THEN
		                      DO;
				   file_nm="&job_end_file_nm.";
				   transfer_to_site_in=1;
				   OUTPUT;
							  END;

		 file_nm="&prefix.file_list_out_for_csv.sas7bdat";
		 transfer_to_site_in=1;
			  OUTPUT;
			END;  /*End of IF last  */
	 RUN;
 
%export_sas_to_dlm_txt(tbl_name_in=&msoc_lib..&prefix.file_list_out_for_csv
                                          ,file_name_out=&msoc_dir.file_list.csv);

		 DATA  _NULL_;	
		   CALL SLEEP(3,1);	
		 RUN;

 %create_empty_file(file_nm=&msoc_dir.files_done.ok);

	  %PUT;
	 %PUT End of DP step: dp_cd=&dp_cd. iter_nb=&iter_nb. step_nb=&step_nb. ;
	  %PUT 	last_iter_in=&last_iter_in.  exit_after_each_iter=&exit_after_each_iter.  end_job_dp_in=&end_job_dp_in.;
	 %PUT syscc=&syscc.;
	 %PUT step_return_cd=&step_return_cd. ; 
     %PUT step_return_msg=&step_return_msg. ;
 	  %PUT;
  
                                      %END; /* End of loop DO UNTIL (&last_iter_in.=1;)	  */
																
  /* 
	When &runid_return_dp_cd. >=2000 OR &step_return_cd. >2000 the error is due to incorrect setup at DP.
	 For example, absent or wrong regression dataset, incorrect directories, etc. 
	In such cases, the program will go straight to the labelendmac_fatal_err  below. 
    No files will be sent to central and no files that came from central
    at the beginning of the current step will be deleted. 
   Since no files will be transfered to central the central site will stay in waiting mode
   as if the program at the DP never started. This allows DP simply
   restart SAS program once the initial problem (e.g. wrong regression dataset) 
   is corrected. 
*/

     %endmac_fatal_err:

  %IF &runid_return_dp_cd. >0 AND &step_return_cd. = 0%THEN
 				  %DO;
		 %LET step_return_cd=&runid_return_dp_cd.;
		 %LET step_return_msg=&runid_return_dp_msg.;
				  %END;

   %IF &runid_return_dp_cd.=0 AND &step_return_cd. >0  %THEN
                   %DO;
		  %LET runid_return_dp_cd=&step_return_cd.;
		 %LET runid_return_dp_msg=&step_return_msg.;
		               %END;

   /*	 20170130: Commented out code below. 
		 Create job_fail.ok only when job fails in controlled way and job_fail.ok is added to the manifest file files_done.csv (see code above).
		  In case, of fatal crash (e.g. typo in the main dir folder's name) exit SAS without triggering next iteration of DMC client
		  This allows simply restarting the SAS program at a DP without affecting the whole process of DR regression  
  */
	/*
		 %IF   &step_return_cd. > 0  %THEN
		          						  %DO;
		 %create_empty_file(file_nm=&msoc_dir.job_fail.ok);
										  %END;
	*/

  %PUT;
	 %PUT End of DP step: dp_cd=&dp_cd. iter_nb=&iter_nb. step_nb=&step_nb. ;
	 %PUT syscc=&syscc.;
	 %PUT runid_return_dp_cd=&runid_return_dp_cd. ; 
     %PUT runid_return_dp_msg=&runid_return_dp_msg. ;
 	  %PUT;
   %PUT ------- End of the macro dp_main ;
   %PUT;

	%MEND;	 